/*
 * Copyright 2015 Luca Beltrame <lbeltrame@kde.org>
 *
 * This file is part of Danbooru Client.
 *
 * Danbooru Client is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Danbooru Client is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Danbooru Client. If not, see <http://www.gnu.org/licenses/>.
 */

#include <KStatusBar>

#include <QAction>
#include <QQuickWidget>
#include <QQmlContext>
#include <QApplication>
#include <QGuiApplication>
#include <QStandardPaths>
#include <QUrl>
#include <QDebug>
#include <QDockWidget>
#include <QQuickItem>
#include <QFileDialog>

#include <kactioncollection.h>
#include <KStandardAction>
#include <KLocalizedString>
#include <KDeclarative/KDeclarative>
#include <KConfigDialog>
#include <KToggleAction>
#include <KDualAction>
#include <KIO/FileCopyJob>
#include <KIO/MimetypeJob>

#include "libdanbooru/danbooruservice.h"
#include "libdanbooru/danboorupost.h"
#include "libdanbooru/danboorupool.h"
#include "model/danboorupostmodel.h"
#include "model/danboorupoolmodel.h"
#include "model/danboorutagmodel.h"
#include "mainwindow.h"
#include "danbooruconnectwidget.h"
#include "danboorusearchwidget.h"
#include "danboorutagwidget.h"
#include "danboorusettings.h"
#include "generalpage.h"
#include "blacklistpage.h"

namespace Danbooru
{

QHash<int, DanbooruPost::Rating> DanbooruMainWindow::ratingMap = {
    {0, DanbooruPost::Safe},
    {1, DanbooruPost::Questionable},
    {2, DanbooruPost::Explicit}
};

DanbooruMainWindow::DanbooruMainWindow(QWidget *parent)
    : KXmlGuiWindow(parent),
      m_view(new QQuickWidget(this)),
      m_model(new DanbooruPostModel(this)),
      m_poolModel(new DanbooruPoolModel(this)),
      m_tagModel(new DanbooruTagModel(this)),
      m_service(new DanbooruService()),
      m_connectWidget(Q_NULLPTR),
      m_searchWidget(new DanbooruSearchWidget(this)),
      m_tagWidget(new DanbooruTagWidget(this)),
      m_tableView(new QTableView(this)),
      m_cache(Q_NULLPTR)
{

    m_service->setParent(this);
    setCentralWidget(m_view);

    qmlRegisterType<Danbooru::DanbooruPost>("DanbooruClient", 1, 0, "DanbooruPost");
    qmlRegisterType<Danbooru::DanbooruService>("DanbooruClient", 1, 0, "DanbooruService");
    qRegisterMetaType<DanbooruPost::Rating>();

    loadSettings();

    m_cache = new KImageCache(qApp->applicationName(), DanbooruSettings::self()->cacheSize());
    m_service->setImageCache(m_cache);
    m_tableView->setModel(m_poolModel);
    m_tagWidget->setModel(m_tagModel);

    // Set up declarative bindings for the QQuickWidget

    m_view->setResizeMode(QQuickWidget::SizeRootObjectToView);
    KDeclarative::KDeclarative declarative;
    declarative.setDeclarativeEngine(m_view->engine());
    declarative.setupBindings();

    auto qmlViewPath = QStandardPaths::locate(QStandardPaths::GenericDataLocation,
                                              qApp->applicationName() +                                             QLatin1String("/danbooruimageview.qml"));

    QQmlContext *ctxt = m_view->rootContext();
    ctxt->setContextProperty("danbooruModel", m_model);
    ctxt->setContextProperty("danbooruService", m_service);
    m_view->setSource(QUrl::fromLocalFile(qmlViewPath));
    m_view->rootObject()->setProperty("poolMode", QVariant(false));

    statusBar()->addPermanentWidget(m_connectWidget);

    setupDockWidgets();

    // then, setup our actions
    setupActions();
    setupGUI(KXmlGuiWindow::ToolBar|Keys|Save|Create|StatusBar, "danbooru-clientui.rc");

    // connections

    connect(m_connectWidget, &DanbooruConnectWidget::accepted, [this]() {

        m_service->setBoardUrl(m_connectWidget->boardUrl());
        if (!m_connectWidget->isAnonymous() && !m_connectWidget->username().isEmpty()
            && !m_connectWidget->password().isEmpty()) {

            m_service->setUserName(m_connectWidget->username());
            m_service->setPassword(m_connectWidget->password());
        }

        actionCollection()->action(QLatin1String("fetch"))->setEnabled(true);
        actionCollection()->action(QLatin1String("find"))->setEnabled(true);
        actionCollection()->action(QLatin1String("poolDownload"))->setEnabled(true);
        actionCollection()->action(QLatin1String("tags"))->setEnabled(true);

        if (DanbooruSettings::self()->autoDownload()) {
            m_view->rootObject()->setProperty("poolMode", QVariant(false));
            m_service->setPostTags(QStringList());
            m_service->getPostList();
            m_service->getTagList(30);
        }
        m_connectWidget->hide();
        statusBar()->hide();

    });

    connect(m_connectWidget, &DanbooruConnectWidget::rejected, [this]() {
        m_connectWidget->hide();
    });

    connect(m_service, &Danbooru::DanbooruService::postDownloaded, m_model,
            &Danbooru::DanbooruPostModel::addPost);
    connect(m_service, &Danbooru::DanbooruService::poolDownloaded, m_poolModel,
            &DanbooruPoolModel::addPool);
    connect(m_service, &Danbooru::DanbooruService::tagDownloaded, m_tagModel,
            &DanbooruTagModel::addTag);
    connect(m_service, &Danbooru::DanbooruService::poolDownloadFinished, [this]() {
        m_tableView->resizeColumnsToContents();
        }
    );

    connect(m_tableView, &QTableView::doubleClicked, [this](QModelIndex index) {
        auto pool = m_poolModel->poolAt(index.row());
        clearModels();
        m_view->rootObject()->setProperty("poolMode", QVariant(true));
        m_service->getPool(pool->id());

    });

    connect(m_searchWidget, &DanbooruSearchWidget::accepted, [this]() {

       QDockWidget* searchDockWidget = findChild<QDockWidget*>(QLatin1String("SearchView"));
       searchDockWidget->hide();

       clearModels();
       m_service->setPostTags(m_searchWidget->selectedTags());
       m_view->rootObject()->setProperty("poolMode", QVariant(false));
       m_service->getPostList();
       m_service->getRelatedTags(m_searchWidget->selectedTags());
    });

    connect(m_searchWidget, &DanbooruSearchWidget::rejected, [this]() {
        QDockWidget* searchDockWidget = findChild<QDockWidget*>(QLatin1String("SearchView"));
        searchDockWidget->hide();
    });

    connect(m_view->rootObject(), SIGNAL(downloadRequested(const QUrl&)), this,
            SLOT(slotHandleDownload(const QUrl&)));

    connect(m_tagWidget, &QListView::doubleClicked, this, &DanbooruMainWindow::searchTag);

}

DanbooruMainWindow::~DanbooruMainWindow()
{
}

void DanbooruMainWindow::loadSettings()
{
    m_service->setBlacklist(DanbooruSettings::self()->tagBlacklist());
    m_service->setMaxPosts(DanbooruSettings::self()->maxPosts());

    m_service->setMaximumAllowedRating(ratingMap.value(DanbooruSettings::self()->maxRating()));

    m_service->setMaxPosts(DanbooruSettings::self()->maxPosts());

    QVector<QUrl> boardsList;

    QStringList::const_iterator it;

    for(it = DanbooruSettings::self()->boards().constBegin();
        it != DanbooruSettings::self()->boards().constEnd();
        ++it) {

        boardsList.append(QUrl::fromUserInput(*it));

    }

    if (!m_connectWidget) {
        m_connectWidget = new DanbooruConnectWidget(boardsList, this);
    } else {
        m_connectWidget->setBoards(boardsList);
    }

    m_tagWidget->setBlackList(DanbooruSettings::self()->tagBlacklist());

}


void DanbooruMainWindow::setupActions()
{

    QAction *connectAction = new QAction(
        QIcon::fromTheme(QLatin1String("document-open-remote")),
        i18n("Connect..."),
        this);

    QAction *fetchAction = new QAction(QIcon::fromTheme(QLatin1String("download")),
                                       i18n("Download"), this);
    KToggleAction *findAction = new KToggleAction(QIcon::fromTheme(QLatin1String("edit-find")),
                                      i18n("Search"), this);
    KToggleAction *poolAction = new KToggleAction(QIcon::fromTheme(QLatin1String("image-x-generic")),
                                                  i18n("Pools"), this);

    KDualAction* tagAction = new KDualAction(i18n("Show tags"), i18n("Hide tags"), this);
    tagAction->setIconForStates(QIcon::fromTheme(QLatin1String("tag")));

    fetchAction->setEnabled(false);
    findAction->setEnabled(false);
    poolAction->setEnabled(false);

    poolAction->setChecked(false);
    findAction->setChecked(false);
    tagAction->setEnabled(false);

    actionCollection()->addAction(QLatin1String("connect"), connectAction);
    actionCollection()->addAction(QLatin1String("fetch"), fetchAction);
    actionCollection()->addAction(QLatin1String("find"), findAction);
    actionCollection()->addAction(QLatin1String("poolDownload"), poolAction);
    actionCollection()->addAction(QLatin1String("tags"), tagAction);

    actionCollection()->setDefaultShortcut(connectAction, KStandardShortcut::Open);
    actionCollection()->setDefaultShortcut(findAction, KStandardShortcut::Find);

    KStandardAction::quit(qApp, SLOT(quit()), actionCollection());
    KStandardAction::preferences(this, SLOT(optionsPreferences()), actionCollection());

    connect(connectAction, &QAction::triggered, this, &DanbooruMainWindow::connectToBoard);
    connect(fetchAction, &QAction::triggered, this, &DanbooruMainWindow::downloadPosts);

    connect(poolAction, &KToggleAction::toggled, [this](bool checked) {
        if (!m_service) {
            return;
        }

        QDockWidget* poolDockWidget = findChild<QDockWidget*>(QLatin1String("PoolView"));

        if (checked) {

            if (m_poolModel->rowCount() == 0) {
                m_service->getPoolList();
            }

            poolDockWidget->show();
            m_tableView->show();


        } else {
            poolDockWidget->hide();
            m_tableView->hide();
        }
    });

    connect(findAction, &KToggleAction::toggled, [this](bool checked) {

        QDockWidget* searchDockWidget = findChild<QDockWidget*>(QLatin1String("SearchView"));

        if (checked) {
            searchDockWidget->show();
            m_searchWidget->show();
        } else {
            searchDockWidget->hide();
            m_searchWidget->hide();
        }

    });

    connect(tagAction, &KDualAction::activeChanged, [this](bool checked) {

        QDockWidget* tagDockWidget = findChild<QDockWidget*>(QLatin1String("TagView"));

        if (checked) {
            tagDockWidget->show();
            m_tagWidget->show();
        } else {
            tagDockWidget->hide();
            m_tagWidget->hide();
        }

    });

}


void DanbooruMainWindow::setupDockWidgets() {

    // Set up PoolWidget

    QDockWidget* poolDockWidget = new QDockWidget(i18n("Pools"), this);
    poolDockWidget->setAllowedAreas(Qt::BottomDockWidgetArea);
    poolDockWidget->setWidget(m_tableView);
    poolDockWidget->setObjectName("PoolView");

    // Prevent the use of winId() when detached, leads to QQuickWidget bugs
    poolDockWidget->setFeatures(QDockWidget::DockWidgetClosable);
    addDockWidget(Qt::BottomDockWidgetArea, poolDockWidget);
    m_tableView->hide();
    poolDockWidget->hide();

    // Find widget
    QDockWidget *searchDockWidget = new QDockWidget(QLatin1String(""), this);
    searchDockWidget->setAllowedAreas(Qt::TopDockWidgetArea);
    searchDockWidget->setWidget(m_searchWidget);
    searchDockWidget->setObjectName("SearchView");

    searchDockWidget->setFeatures(QDockWidget::NoDockWidgetFeatures);
    addDockWidget(Qt::TopDockWidgetArea, searchDockWidget);
    // No title wanted
    searchDockWidget->setTitleBarWidget(new QWidget(this));

    searchDockWidget->hide();
    m_searchWidget->hide();

    QDockWidget *tagDockWidget = new QDockWidget(QLatin1String("Tags"), this);
    tagDockWidget->setAllowedAreas(Qt::RightDockWidgetArea);
    tagDockWidget->setWidget(m_tagWidget);
    tagDockWidget->setObjectName("TagView");
    tagDockWidget->setFeatures(QDockWidget::DockWidgetClosable);
    addDockWidget(Qt::RightDockWidgetArea, tagDockWidget);

//     tagDockWidget->hide();
//     m_tagWidget->hide();

    // Connections

    connect(poolDockWidget, &QDockWidget::visibilityChanged, [this](bool visible) {
        actionCollection()->action(QLatin1String("poolDownload"))->setChecked(visible);
    });

    connect(searchDockWidget, &QDockWidget::visibilityChanged, [this](bool visible) {
        actionCollection()->action(QLatin1String("find"))->setChecked(visible);
    });

    connect(tagDockWidget, &QDockWidget::visibilityChanged, [this](bool visible) {
        qobject_cast<KDualAction*>(actionCollection()->action(QLatin1String("tags")))->setActive(visible);
    });

}

void DanbooruMainWindow::connectToBoard()
{

    if (!m_view) {
        return;
    }


    clearModels();
    m_poolModel->clear();
    m_service->reset();
    statusBar()->show();
    m_connectWidget->show();

}

void DanbooruMainWindow::downloadPosts()
{
    if (!m_service) {
        return;
    }

    qDebug() << "In download!";

    clearModels();
    m_service->setPostTags(QStringList());
    m_view->rootObject()->setProperty("poolMode", QVariant(false));
    m_service->getPostList();
    m_service->getTagList(30);

}

void DanbooruMainWindow::optionsPreferences()
{
    KConfigDialog* dialog = new KConfigDialog(this, "danboorusettings",
                                              DanbooruSettings::self());

    dialog->addPage(new GeneralPage(DanbooruSettings::self(), this), i18n("General"),
        "table");
    dialog->addPage(new BlacklistPage(DanbooruSettings::self(), this), i18n("Tag blacklist"),
                    "configure");
    connect(dialog, &KConfigDialog::settingsChanged, this, &DanbooruMainWindow::loadSettings);
    dialog->show();
}

void DanbooruMainWindow::slotHandleDownload(const QUrl &url) {

    QFileDialog saveDialog(this);
    saveDialog.setFileMode(QFileDialog::AnyFile);
    saveDialog.setWindowTitle(i18n("Save file"));
    saveDialog.setAcceptMode(QFileDialog::AcceptSave);

    KIO::MimetypeJob* mimeJob = KIO::mimetype(url, KIO::HideProgressInfo);

    if (mimeJob->exec()) {
        saveDialog.setMimeTypeFilters(QStringList(mimeJob->mimetype()));
    }  else {
        saveDialog.setNameFilter(i18n("Images (*.png *.gif *.jpg);;All files (*.*)"));
    }

    // TODO: Remember last user directory - settings?

    saveDialog.setConfirmOverwrite(true);
    saveDialog.setDirectory(QDir::homePath());
    saveDialog.selectFile(url.fileName());

    if (saveDialog.exec()) {
        QUrl localFile = saveDialog.selectedUrls().at(0);

        if (localFile.isEmpty()) {
            return;
        }

        KIO::FileCopyJob *job = KIO::file_copy(url, localFile, -1, KIO::DefaultFlags);

        connect(job, &KIO::Job::result, [this](KJob  *job) {
            if (job->error()) {
                return;
            }
            // TODO: Tagging
        });

    }

}

void DanbooruMainWindow::searchTag(const QModelIndex &index)
{

    if (!index.isValid()) {
        return;
    }

    QString tagName = m_tagModel->itemAt(index.row())->name();
    QStringList tags = {tagName};
    clearModels();
    m_view->rootObject()->setProperty("poolMode", QVariant(false));
    m_service->getRelatedTags(tags);
    m_service->setPostTags(tags);
    m_service->getPostList();

}

void DanbooruMainWindow::clearModels()
{
    m_model->clear();
    m_tagModel->clear();
//     m_poolModel->clear();
    m_service->reset();

}

} // namespace Danbooru
